;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Software License Agreement (BSD License)
;; 
;; Copyright (c) 2008, Willow Garage, Inc.
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with 
;; or without modification, are permitted provided that the 
;; following conditions are met:
;;
;;  * Redistributions of source code must retain the above 
;;    copyright notice, this list of conditions and the 
;;    following disclaimer.
;;  * Redistributions in binary form must reproduce the 
;;    above copyright notice, this list of conditions and 
;;    the following disclaimer in the documentation and/or 
;;    other materials provided with the distribution.
;;  * Neither the name of Willow Garage, Inc. nor the names 
;;    of its contributors may be used to endorse or promote 
;;    products derived from this software without specific 
;;    prior written permission.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND 
;; CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED 
;; WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
;; PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
;; COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
;; INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
;; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
;; CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
;; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
;; OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
;; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
;; DAMAGE.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :gtest-adapter)


;;; Functions that allow using rostest which is gtest based.
;;; allows writing test results to gtest compatible files
;;;
;;; Wrapping any test framework for this is a 2 step process.
;;; provide a function that runs all test of a testsuite and returns
;;; some result. The also provide a function which transforms
;;; that result into a gtestsuite structure. Those steps are separated
;;; as standard-output of tests is captured but not of transforming.
;;; However you are free to return a gtestsuite as the result of your
;;; suite-fun and use 'identity as transfrom function.

;; the following structures capture the information actually used by gtest, which is a subset of the junitreport format.
(defstruct gtestfailure
  type ;; string
  message ;; string
  )

(defstruct gtestcase
  classname ;; string
  name ;; string
  time ;; float
  failures ;; list of gtestfailure
  )

(defstruct gtestsuite
  name ;; string
  time ;; float
  testcases ;; list of gtestcase
  sysout ;; string
  syserr ;; string
  )

;; Possible args with which a gtest-executable can be called
(defstruct gtestargs
  output-xml;; string
  ;; TODO: more args
  )



(defun ensure-float (var)
  "transforms whatever it gets to a float, tries to be robust"
  (if (null var) 0.0
    (coerce 
     (if (numberp var) 
       var 
       (ensure-float (read-from-string var)))
     'float)))

(defun ensure-int (var)
  "transforms whatever it gets to an integer, tries to be robust"
  (if (null var) 0
    (coerce 
     (if (numberp var) 
       var 
       (ensure-int (read-from-string var)))
     'integer)))

;; stolen from tbnl (http://weitz.de/tbnl/)
(defun escape-for-html (string)
  "Escapes the characters #\\<, #\\>, #\\', #\\\", and #\\& for HTML output."
  (with-output-to-string (out)
    (with-input-from-string (in (format nil "~a" string))
      (loop for char = (read-char in nil nil)
            while char
            do (case char
                 ((#\<) (write-string "&lt;" out))
                 ((#\>) (write-string "&gt;" out))
                 ((#\") (write-string "&quot;" out))
                 ((#\') (write-string "&#039;" out))
                 ((#\&) (write-string "&amp;" out))
                 (otherwise (write-char char out)))))))

(defun escape-quotes (string)
  "Escapes the character  #\\\", XML atribute values."
  (with-output-to-string (out)
    (with-input-from-string (in (format nil "~a" string))
      (loop for char = (read-char in nil nil)
            while char
            do (case char
                 ((#\") (write-string "&quot;" out))
                 (otherwise (write-char char out)))))))

(defun escape-cdata (string)
  "in cdata anything is allowed except ']]>', which must be prefixed by ']]><![CDATA[' here."
  (let ((state 0))
    (with-output-to-string (out)
      (with-input-from-string (in (format nil "~a" string))
        (loop for char = (read-char in nil nil)
              while char
              do (case char
                   ((#\])
                     (write-char char out)
                     (when (< state 2  (incf state))))
                   ((#\>)
                    (if (= state 2)
                      ;; this splits the cdata blocks, which seems to be the common way of dealing with cdata
                      (write-string "]]><![CDATA[]" out)
                      (write-char char out))
                    (setf state 0))
                   (otherwise
                    (write-char char out)
                    (setf state 0))))))))

(defun serialize-test-result (stream gtestcase)
  "Writes to stream an XML string representing the testcase as gtest does. Failures is a list of list-pairs where 1st is type name, 2nd is message"
  ;; status not used by gtest nor rt
  (format stream "<testcase classname=\"~a\" name=\"~a\" status=\"run\" time=\"~a\">~%"
          (escape-quotes (gtestcase-classname gtestcase))
          (escape-quotes (gtestcase-name gtestcase))
          (ensure-float (gtestcase-time gtestcase)))
  (dolist (failure (gtestcase-failures gtestcase))
    (format stream "<failure type=\"~a\">~a</failure>\~%"
            (escape-quotes (gtestfailure-type failure))
            (escape-for-html (gtestfailure-message failure))))
  (format stream "</testcase>~%"))

;; TODO escape for xml, checkl types, format time
(defun serialize-testsuite (stream gtestsuite)
  "Writes to stream an XML string representing the testsuite as gtest does."
  ;; error not used by gtest, nor rt
  (multiple-value-bind (n-tests n-failures) (analyze-results gtestsuite)
           (format stream "<testsuite name=\"~a\" tests=\"~a\" failures=\"~a\" errors=\"0\" time=\"~a\">~%"
                   (escape-quotes (gtestsuite-name gtestsuite))
                   (ensure-int n-tests)
                   (ensure-int n-failures)
                   (ensure-float (gtestsuite-time gtestsuite))))
  (dolist (result (gtestsuite-testcases gtestsuite))
    (serialize-test-result
     stream
     result))
  (format stream "<system-out><![CDATA[~a]]></system-out>~&" (escape-cdata (gtestsuite-sysout gtestsuite)))
  (format stream "<system-err><![CDATA[~a]]></system-err>~&" (escape-cdata (gtestsuite-syserr gtestsuite)))
  (format stream "</testsuite>~%"))


(defun analyze-results (gtestsuite)
  "counts tests and failures"
  (let ((n-tests 0)
        (n-failures 0))
    (dolist (gtestcase (gtestsuite-testcases gtestsuite))
      (if (null (gtestcase-failures gtestcase))
        (incf n-tests)
        (incf n-failures)))
    (values n-tests n-failures)))


;;;; Public functions to run suites

(defun run-suite (suite-fun arg-struct)
  "funcalls suite-fun with arg-struct and returns list (result standard-output and error-output)"
  (let (sysout
         syserr
         result)
    (setf sysout
          (with-output-to-string (outstream)
            (setf syserr
                  (with-output-to-string (errstream)
                    (let ((*standard-output* outstream)
                          (*error-output* errstream))
                      (setf result (funcall suite-fun arg-struct)))))))
    (list result sysout syserr)))

;; the default behavior for tests
(defun run-suite-with-output (suite-fun arg-struct &key (no-output nil))
  "Funcalls suite-fun with arg-struct as arg and writes result to stream. Test-fun should run without args and return a gtestsuite structure."
  (destructuring-bind (gtestsuite
                       sysout
                       syserr)
      (run-suite suite-fun arg-struct)
    (unless no-output
    ;; output supressed messages
      (format t "~a~%" sysout)
      (format *error-output* "~a~%" syserr))
    (list gtestsuite sysout syserr)))


;; best for functional self-tests
(defun run-suite-write-to-stream (stream suite-fun arg-struct transform-fun &key (no-output nil))
  "Funcalls suite-fun with arg-struct as arg and writes result to
stream. Test-fun should run without args and return a gtestsuite
structure."
  (let* ((result (run-suite-with-output suite-fun arg-struct :no-output no-output))
         (gtestsuite (apply transform-fun result)))
      (serialize-testsuite stream gtestsuite)))


(defun run-suites-write-to-file (filename suite-fun-list arg-struct transform-fun &key (no-output nil))
  "Funcalls all functions in suite-fun-list with arg-struct as arg, calls transform-fun
on (result sysout syserr) and writes results as XML to
filename. every function in suite-fun should run without args and return a gtestsuite."
  (if (null filename)
    (dolist (suite-fun suite-fun-list)
        (run-suite-write-to-stream nil suite-fun arg-struct transform-fun :no-output no-output))
    (with-open-file (stream
                     filename
                     :direction :output
                     :if-exists :supersede)
      (format stream "<testsuites>~%")
      (dolist (suite-fun suite-fun-list)
        (run-suite-write-to-stream stream suite-fun arg-struct transform-fun :no-output no-output))
      (format stream "</testsuites>~%"))))


;; the most convenient function to use
(defun run-suites-write-gtest-file (posix-args suite-fun-list transform-fun &key (no-output nil))
  "Funcalls every function in suite-fun-list with arg-struct, writes result to file as given by posix-args in gtest format, else output to stdout"
  (if (find-if (lambda (x) (search "--help" x)) posix-args)
    (format t "Usage: <executable> --gtest_output=xml:<filename> ~%")
    (let ((arg-struct (get-gtest-args posix-args)))
      (when (null (gtestargs-output-xml arg-struct))
        (format *error-output* "No --gtest_output=xml:filename given in args ~a~%" (cdr posix-args)))
      (run-suites-write-to-file (gtestargs-output-xml arg-struct)
                                suite-fun-list
                                arg-struct
                                transform-fun
                                :no-output no-output))))

;;;;;;; gtest Command line arg parsing

(defun get-gtest-args (args)
  "Parses args looking for occurence of keys in args and returns a struct with all detected args"
  (let ((result (make-gtestargs)))
    (dolist (arg args)
      (let ((index-filename (search "--gtest_output=xml:" arg)))
        (when (and index-filename (= 0 index-filename))
          (setf
           (gtestargs-output-xml result)
           (subseq arg (length "--gtest_output=xml:"))))))
    result))